#include <iostream>
#include <verilated.h>
#include "VRAS.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t max_time = 200;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VRAS* top = new VRAS();

  top->clk = 0;
  top->res_n = !0;

  top->lock = 0;
  top->unlock = 0;
  
  top->in_addr_val = 0;
  top->in_addr = 0xFEDCBABCDEF0;
    
  top->rd = 0;

  while(!Verilated::gotFinish()){

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(!top->clk){
    
      // Toggle control signals on an edge that doesn't correspond
      // to where the controls are sampled; in this example we do
      // this only on a negedge of clk, because we know
      // reset is not sampled there.
      if ( main_time > 0 && main_time <= 4 ) {
	top->res_n = !1;  //   Assert reset
      } else {
	top->res_n = !0;  // Deassert reset
      }

      if( main_time>4 && main_time<=6 ){
	// unlock RAS
	top->key = 0xDEADBEEF;
	top->unlock = 1;
      }

      	if (main_time>6 && main_time<=floor(max_time/4)) {
	  top->unlock = 0;
	  top->key = 0x00000000;
	  if (0==main_time%4) {
	    // increment or wrap address
	    if(top->in_addr==0xFFFFFFFFFFFF) {
	      top->in_addr=0x000000000000;
	    } else {
	      top->in_addr += 0x00000000000F;
	    }
	  }
	  if (0==main_time%4 ) {
	  // write to stack
	    top->in_addr_val = 1;	    
	  } else {
	    top->in_addr_val = 0;
	  }
	}

	if (main_time>floor(max_time/4) && main_time<=floor(max_time/2)) {
	  // read from stack
	  top->in_addr_val = 0;
	  if (0==main_time%4) {
	    top->rd = 1;
	  } else {
	    top->rd = 0;
	  }
	}
      
	if (main_time>floor(max_time/2) && main_time<=floor(4*max_time/6)) {
	  
	  // lock stack
	  if(main_time==floor(max_time/2)+2) top->lock = 1; else top->lock = 0;

	  // and read from locked stack
	  if (0==main_time%4) {
	    top->rd = 1;
	  } else {
	    top->rd = 0;
	  }
	}
      
	if (main_time>floor(4*max_time/6) && main_time<=floor(5*max_time/6)) {

	  // lock stack
	  if(main_time==floor(max_time/2)+2) top->lock = 1; else top->lock = 0;

	  // toggle in_addr
	  if (0==main_time%4) {
	    // increment or wrap address
	    if(top->in_addr==0xFFFFFFFFFFFF) {
	      top->in_addr=0x000000000000;
	    } else {
	      top->in_addr += 0x00000000000F;
	    }
	  }

	  // and write to locked stack
	  if (0==main_time%4) {
	    top->in_addr_val = 1;
	  } else {
	    top->in_addr_val = 0;
	  }
	}
	
	if (main_time>floor(5*max_time/6) && main_time<=floor(max_time)) {

	  // unlock stack
	  if(main_time==floor(5*max_time/6)+2){
	    top->lock = 0;
	    top->unlock = 1;
	    top->key = 0xDEADBEEF;
	  } else {
	    top->lock = 0;
	    top->unlock = 0;
	    top->key = 0x00000000;
	  }

	  if (0==main_time%4) {
	    // increment or wrap address
	    if(top->in_addr==0xFFFFFFFFFFFF) {
	      top->in_addr=0x000000000000;
	    } else {
	      top->in_addr += 0x00000000000F;
	    }
	  }

	  // concurrent read write access
	  if (0==main_time%4) {
	    top->in_addr_val = 1;
	    top->rd = 1;
	  } else {
	    top->in_addr_val = 0;
	    top->rd = 1;
	  }
	}      

      if(main_time>=max_time){	
	top->in_addr_val=1; top->lock=1; top->unlock=1; top->res_n=0;
      }
	
    }
       
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    if(!top->clk){
      // Read outputs
      VL_PRINTF( "[%" VL_PRI64 "d] clk=%x res_n=%x lock=%x ( unlock=%x key=%x ) ( in_addr_val=%x in_addr=%12x ) rd=%x => locked=%x   ( out_addr_val=%x, out_addr=%12x )  ( empty=%x full=%x )  ras_write_active=%x\n",
		 main_time,
		 top->clk, top->res_n,
		 top->lock, top->unlock, top->key,
		 top->in_addr_val, top->in_addr, top->rd,
		 top->locked, top->out_addr_val, top->out_addr, top->empty, top->full, top->ras_write_active );             
      }
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
